/*
 * Le Lann, Chang and Roberts leader election algorithm for a ring of
 * processes
 *
 * NOTEs(PT):
 * This solution contains a 'simplified' random initialization procedure.
 * By adding some random base value j to the generated unique ids, the
 * nodes can not individually compute who is the leader without exchanging
 * some messages.
 *
 * If you try to lower the value of J_BOUND, you can see that the verifier
 * is progressively faster, due to the reduction of search space size. :-)
 *
 * author: unknown
 *
 */

#define N 5
#define J_BOUND (255 - N - 1) // Try with: 100, 50, 10
#define BUFSIZE 10

mtype = { candidate, leader };
chan c[N] = [BUFSIZE] of { mtype, byte };

byte num_leaders = 0;

proctype node(chan prev, next; byte my_id)
{
    show byte in_id;
    show byte leader_id;

    // init: pass the node id to its neighbour
    next!candidate(my_id);

    do
        :: prev?candidate(in_id) ->
            if
                :: in_id > my_id ->
                    next!candidate(in_id);
                :: in_id < my_id ->
                    skip;
                :: else ->
                    printf("I am the leader! (pid: %d, id: %d)\n", _pid, my_id);
                    num_leaders++;
                    next!leader(my_id);
            fi
        :: prev?leader(leader_id) ->
            if
                :: leader_id > my_id ->
                    next!leader(leader_id);
                    break;
                :: else ->
                    assert(my_id == leader_id);
                    break;
            fi
    od;
    assert(num_leaders == 1); // NOTE: if more than one node
                              // declares itself as leader, then
                              // this assertion will fail for
                              // at least one node!
}

init {
    byte proc;
    byte i, j;

    // here atomic is 'used' to delay all the nodes, so that they
    // start executing after the initialization has finished
    atomic {
        chan prev, next;
        byte id;
        // pick up random j
        j = 0;
        do
            :: j < J_BOUND ->
                j++;
            :: true ->
                break;
        od;
        // pick up random i
        i = 1;
        do
            :: i < N -> i++;
            :: true -> break;
        od;
        do
            :: proc < N ->
                prev = c[proc];
                next = c[(proc+1)%N];
                id = (N+i-proc)%N + j;
                printf("Starting process with id %d.\n", id);
                run node(prev, next, id);
                proc++;
            :: else ->
                break;
        od
    }
}
